%% DEMO_DENOISING - Demonstrate MMSE-based blind denoising with integrated noise estimation
% 
%  Author:  Uwe Schmidt, Department of Computer Science, TU Darmstadt
%  Contact: uwe.schmidt@gris.tu-darmstadt.de

% Copyright 2011 TU Darmstadt, Darmstadt, Germany.
% $Id: demo_denoising.m 246 2011-05-30 12:32:40Z uschmidt $

function demo_denoising
  [prev_dir, base_dir] = adjust_path;
  
  fprintf('Demo of blind denoising with integrated noise estimation.\n')
  choice = input('(1) for pairwise MRF, or (2) for 3x3 FoE with 8 filters: ');
  if choice == 1, mrf = learned_models.cvpr_pw_mrf;  fprintf('Using pairwise MRF\n'); else
                  mrf = learned_models.cvpr_3x3_foe; fprintf('Using 3x3 FoE\n'); end
  
  sigma = 25*rand; % choose noise std
  img_names = {'barbara.png', 'boat.png', 'fingerprint.png', 'house.png', 'lena.png', 'peppers256.png'};
  idx = 2; % choose index of img_names
  img_clean = double(imread(sprintf('images/%s', img_names{idx})));
  sc = 1/4; % choose scale (1 => original size)
  img_clean = imresize(img_clean, sc);
  img_noisy = img_clean + sigma * randn(size(img_clean));
  
  % blind denoising with noise estimation
  [img_denoised, sigma_est] = deblur_mmse(mrf, img_noisy, 1, img_clean, sigma, [], true, true, true);
  
  adjust_path(prev_dir, base_dir);
end

function [prev_dir, base_dir] = adjust_path(prev_dir, base_dir)
  if nargin == 2
    % restore working directory
    % restoring the path sometimes confuses MATLAB when running the code again ("clear classes" helps)
    cd(prev_dir); % rmpath(base_dir); 
    warning('on','MATLAB:mir_warning_maybe_uninitialized_temporary')
  else
    % save working directory and go to correct directory
    prev_dir = pwd; file_dir = fileparts(mfilename('fullpath')); cd(file_dir);
    last = @(v) v(end); base_dir = file_dir(1:last(strfind(file_dir, filesep))-1);
    % add base directory to path
    addpath(base_dir);
    warning('off','MATLAB:mir_warning_maybe_uninitialized_temporary')
  end
end